"""API endpoints for Connect-N API"""

# Python module(s)
import json

# Flask module(s)
from flask import request, jsonify, g

# App module(s)
from .app import app, db
from .db_model import User
from .auth import token_auth
from .utility import compile_response
from .error import bad_request

# Project module(s)
from ..connect_n import ConnectNGame


@app.route("/", methods=["GET"])
@app.route("/test", methods=["GET"])
@app.route("/index", methods=["GET"])
def index():
    """API endpoint for index/homepage/testing"""
    return compile_response(description="Test Request")


@app.route("/new-user", methods=["POST"])
def create_user():
    """API endpoint to create new user"""
    data = request.get_json() or {}
    if "username" not in data or "email" not in data or "password" not in data:
        return bad_request("Must include username, email and password fields", 400)
    if User.query.filter_by(username=data["username"]).first():
        return bad_request("Please use a different username", 400)
    if User.query.filter_by(email=data["email"]).first():
        return bad_request("Please use a different email", 400)
    user = User()
    user.from_dict(data, new_user=True)
    db.session.add(user)
    db.session.commit()
    response = jsonify(compile_response(description=user.to_dict()))
    response.status_code = 201
    return response


@app.route("/get-user/<int:id>", methods=["GET"])
@token_auth.login_required
def get_user(id):
    """API endpoint to retrive specefic user info from id"""
    u = User.query.get(id).to_dict() or {}
    response = jsonify(compile_response(user=u))
    response.status_code = 201
    return response


@app.route("/new-game", methods=["GET"])
@token_auth.login_required
def start_new_game():
    """"""
    user = User.query.get(g.current_user.id)
    if user.game != "Empty":
        return jsonify(
            compile_response(
                description="Previous game already in progress, try /delete to remove previous game"
            )
        )
    game = ConnectNGame()
    g.current_user.game = json.dumps(game.to_dict(), default=lambda x: x.__dict__)
    db.session.commit()
    response = jsonify(
        compile_response(description="New game created", game=g.current_user.game)
    )
    response.status_code = 201
    del game
    return response


@app.route("/get-game", methods=["GET"])
@token_auth.login_required
def get_game():
    """API endpoint to retrive current game"""
    user = User.query.get(g.current_user.id)
    game = user.game
    response = jsonify(compile_response(description="Game ended", game=game))
    response.status_code = 201
    return response


@app.route("/make-game/<int:move>", methods=["GET"])
@token_auth.login_required
def make_move(move):
    """API endpoit to make move in the game

    Args:
        move (int): Player move"""
    user = User.query.get(g.current_user.id)
    game_dict = json.loads(user.game)
    game = ConnectNGame()
    game.from_dict(game_dict)
    game.make_move(move, 2)
    user.game = json.dumps(game.to_dict())
    db.session.commit()
    response = jsonify(compile_response(description="Game ended", game=game.to_dict()))
    response.status_code = 201
    return response


@app.route("/end-game", methods=["GET"])
@token_auth.login_required
def end_game():
    """API endpoint to delete current game if any"""
    user = User.query.get(g.current_user.id)
    game = user.game
    user.game = "Empty"
    db.session.commit()
    response = jsonify(compile_response(description="Game ended", game=game))
    response.status_code = 201
    return response



"""Database Model(s) for Connect-N API"""

# Python module(s)
import os
import base64
from datetime import datetime, timedelta

from werkzeug.security import generate_password_hash, check_password_hash

# pylint: disable=no-member

from .app import db


class User(db.Model):
    """User table defination"""

    id = db.Column(db.Integer, primary_key=True)
    username = db.Column(db.String(64), nullable=False)
    password_hash = db.Column(db.String(128))
    email = db.Column(db.String(120), index=True, unique=True)
    last_request = db.Column(db.DateTime, default=datetime.utcnow, nullable=False)
    token = db.Column(db.String(32), index=True, unique=True)
    token_expiration = db.Column(db.DateTime, default=datetime.utcnow)
    game = db.Column(db.Text, default="Empty")

    def set_password(self, password):
        """Method used to set user password

        Args:
            password (str): Password set as password hash"""
        self.password_hash = generate_password_hash(password)

    def check_password(self, password):
        """Method to check user password

        Args:
            password (str): Password to check against

        Returns:
            bool : True if correct password"""
        return check_password_hash(self.password_hash, password)

    def from_dict(self, data, new_user=False):
        """Method to read user from dictioary

        Args:
            data (dict): Must contain username, email and password fields
            new_user (bool): If new user registrations"""
        for field in ["username", "email"]:
            if field in data:
                setattr(self, field, data[field])
        if new_user and "password" in data:
            self.set_password(data["password"])

    def to_dict(self, include_email=False):
        """Provides dictionary representation of user

        Args:
            include_email: Wheather to include email

        Returns:
            dict : dictionary representation"""
        data = {
            "id": self.id,
            "username": self.username,
            "last_request": self.last_request.isoformat() + "Z",
        }
        if include_email:
            data["email"] = self.email
        return data

    def get_token(self, expires_in=3600):
        """Token generator

        Args:
            expires_in (int): Seconds in which token expires

        Returns:
            str : API token for the user"""
        now = datetime.utcnow()
        if self.token and self.token_expiration > now + timedelta(seconds=60):
            return self.token
        self.token = base64.b64encode(os.urandom(24)).decode("utf-8")
        self.token_expiration = now + timedelta(seconds=expires_in)
        db.session.add(self)
        return self.token

    def revoke_token(self):
        """Method to revoke API token"""
        self.token_expiration = datetime.utcnow() - timedelta(seconds=1)

    @staticmethod
    def check_token(token):
        """Static method for token search and verification

        Args:
            token (str): Token to check

        Reurns:
            User : User to whom supplied token belongs to"""
        user = User.query.filter_by(token=token).first()
        if user is None or user.token_expiration < datetime.utcnow():
            return None
        return user



"""Environment variable(s) for Connect-N API"""

MINIMUM_REQUEST_DURATION = 5

API_DB_NAME = "/api.db"

RESPONSE_TEMPLATE = {
    "@author": "Kartikei Mittal",
    "project_link": "https://github.com/Kartikei-12/Connect-N",
    "description": "",
}



"""Utility method(s) for Connect-N API"""

import json

from .env import RESPONSE_TEMPLATE


def compile_response(*args, **kwargs):
    """Function to compile response

    Returns:
        str : JSON response to API call"""
    response = RESPONSE_TEMPLATE.copy()
    for key, value in kwargs.items():
        response[str(key)] = value
    return json.dumps(response)



"""player.py file

Contains player module"""

# Python module(s)
from itertools import count


class Player:
    """Player class

    Args:
        name (str): Name of the player"""

    _ids = count(2)

    def __init__(self, name=""):
        """Instantiate method"""
        self.name = name
        self.p_id = next(self._ids)

    def get_move(self):
        """Simple method to ask for player move

        Returns:
            int : Player input, None if invalid input"""
        msg = "Player {0} make your move: ".format(self.name)
        try:
            col = int(input(msg)) - 1
        except ValueError as e:
            print(e, "Aborting current turn, moving ahead.")
            return None
        else:
            return col

    def __str__(self):
        """String representation of class object"""
        return "<class 'Player'> {0} {1} ".format(self.p_id, self.name)



"""Default and environment variables for this project."""

N = 4
ROWS = 6
COLUMNS = 7

OFFSET = -2 * (10 ** 3)
UNIT_SCORE = 10 ** 3

LARGE_VALUE = 10 ** 12
SMALL_VALUE = -10 ** 12

NEURAL_NET_param = {"name": "Neural Network parameters"}



"""Data preprocessing for neural network trainning"""

# Python module(s)
import numpy as np

from ..env import ROWS, COLUMNS


class Preprocess:
    """Pre-processing class for dataset

    Args:
        dataset (str): PATH to datset(in .npy format)

    Raises:
        TypeError: when incorrect path to dataset file is provided."""

    def __init__(self, dataset=""):
        """Instantiate method"""
        if not isinstance(dataset, str) or not dataset.endswith(".npy"):
            raise TypeError(
                "Name of dataset(with path) in .npy format need to be provided"
            )
        self.X, self.y = None, None
        self.dataset_name = dataset
        self.dataset = np.load(dataset, allow_pickle=True)

    def fit(self):
        """Fit method for data preprocessing"""
        X, y = list(), list()
        new_dataset = list()
        encode = {
            0: np.array([0, 0, 0], dtype=int),
            1: np.array([0, 1, 0], dtype=int),
            2: np.array([0, 0, 2], dtype=int),
        }
        min_game_len = min(i.shape[0] for i in self.dataset)
        for game in self.dataset:
            new_game = list()
            for frame in game:
                new_frame = list()
                for row in frame:
                    new_row = list()
                    for ele in row:
                        new_row.append(encode[int(ele)])
                    new_frame.append(np.array(new_row))
                new_game.append(np.array(new_frame))
            new_dataset.append(np.array(new_game))
        self.dataset = np.array(new_dataset)
        for game in self.dataset:
            for i in range(min_game_len, game.shape[0]):
                X.append(game[i - min_game_len : i]), y.append(game[i])
        self.X, self.y = np.array(X), np.array(y)

    def transform(self):
        """Provides seprated and properly encoded X, y from dataset

        Returns:
            numpy.ndarray : X
            numpy.ndarray : y"""
        return self.X, self.y

    def __str__(self):
        """str method"""
        return "<class Preprocess: {}>".format(self.dataset_name)



"""@author: Kartikei Mittal

__init__.py file connect_n module"""

__author__ = "Kartikei Mittal"
__version__ = "0.1d"
__email__ = "kartikeimittal@gmail.com"

from .connect_n import ConnectNGame
from .player import Player
from .ai_nn import Preprocess, GenerateData
from .env import *

import os

FILE_PATH = os.path.dirname(os.path.realpath(__file__)) + "/version.txt"

try:
    version = str(int(open(FILE_PATH, "r").read()) + 1)
    open(FILE_PATH, "w").write(version)
    del version
except (FileNotFoundError, ValueError):  # version.txt missing
    with open(FILE_PATH, "w+") as f:
        f.write("1")
        f.close()



"""auth.py file for Connect-N game API"""

# Flask module(s) and it's extentions
from flask import g
from flask_httpauth import HTTPBasicAuth, HTTPTokenAuth

# User module(s)
from .db_model import User
from .error import bad_request

token_auth = HTTPTokenAuth()
basic_auth = HTTPBasicAuth()


@basic_auth.verify_password
def verify_password(username, password):
    """API method used to verify password

    Args:
        username (str): Username
        password (str): Password for login

    Returns:
        bool : If valid password for supplied username"""
    user = User.query.filter_by(username=username).first()
    if user is None:
        return False
    g.current_user = user
    return user.check_password(password)


@basic_auth.error_handler
def basic_auth_error():
    """Error Handaler for Authorization Error"""
    return bad_request("Authorization Error", 401)


@token_auth.verify_token
def verify_token(token):
    """API token verification method"""
    g.current_user = User.check_token(token) if token else None
    return g.current_user is not None


@token_auth.error_handler
def token_auth_error():
    """Error Handaler for Authentication Error"""
    return bad_request("Authentication Error", 401)



"""API tokens handaling"""

from flask import jsonify, g

from .app import app, db
from .auth import basic_auth, token_auth
from .utility import compile_response


@app.route("/tokens", methods=["POST"])
@basic_auth.login_required
def get_token():
    """API endpoint to generate new token"""
    token = g.current_user.get_token()
    db.session.commit()
    return jsonify(compile_response(description={"token": token}))


@app.route("/tokens", methods=["DELETE"])
@token_auth.login_required
def revoke_token():
    """API endpoint to revoke user token"""
    g.current_user.revoke_token()
    db.session.commit()
    return "", 204



"""Utility class for pygame

Note:
    In this files pylint configured to ignore 'no member' because of C implementation of several pygame members like constants MOUSEBUTTONDOWN and methods like init()"""

# Python module(s)
import math
import pygame
import random

# Environment Variables
from .env import ROWS, COLUMNS

# Colours
BLUE = (0, 0, 255)
BLACK = (0, 0, 0)
COLOR = {"RED": (255, 0, 0), "YELLOW": (255, 255, 0), "GREEN": (0, 255, 0)}

C_LIST = list(value for key, value in COLOR.items())
C_LIST.insert(0, BLACK)

# Some constants
SQUARESIZE = 90
RADIUS = int(SQUARESIZE / 2 - 5)

# Disabling pylint because of partial implementation of pygame in C, which is not recognizable by pylint
# pylint: disable=no-member


class PygameUtility:
    """Utility class for pygame module

    Note:
        Docker does nor support pygame will automatically switch to command line.

    Args:
        r (int): Rows of game
        c (int): Columns of game"""

    def __init__(self, r=ROWS, c=COLUMNS):
        """Instantiate Method"""
        pygame.init()
        self.num_col = c
        self.num_rows = r
        self.width = c * SQUARESIZE
        self.height = (r + 1) * SQUARESIZE
        try:
            self.screen = pygame.display.set_mode((self.width, self.height))
        except pygame.error as e:  # Expected inside docker.
            print(e, "\nSwitching to command line.")
            raise EnvironmentError("Vedio device not found.")
        pygame.display.update()

    def draw_black_rec(self):
        """Draw empty rectangles."""
        pygame.draw.rect(self.screen, BLACK, (0, 0, self.width, SQUARESIZE))

    def draw_player_coin(self, p_id, event):
        """Draw player coin

        Args:
            p_id (int): Player ID
            event (pygame.event): Event of motion of coin."""

    def blit(self, msg, p_id):
        """Finising game.

        Args:
            msg (str): Message to display
            p_id (int): Winnig player p_id"""
        size = int((self.num_rows * SQUARESIZE * 1.5) / len(msg))
        label = pygame.font.SysFont("monospace", size).render(msg, 1, C_LIST[p_id])
        self.screen.blit(label, (40, 10))

    def draw(self, board):
        """Draw basic board for the game.

        Args:
            board (numpy.ndarray): 2-D numpy array representing the board"""
        self.screen.fill(BLUE, (0, SQUARESIZE, self.width, self.height))
        for c in range(self.num_col):
            for r in range(self.num_rows):
                pygame.draw.circle(
                    self.screen,
                    C_LIST[board[r][c]],
                    (
                        int(c * SQUARESIZE + SQUARESIZE / 2),
                        self.height - int(r * SQUARESIZE + SQUARESIZE / 2),
                    ),
                    RADIUS,
                )
        pygame.display.update()

    def play(self, board, players, is_valid_move, make_move, is_winning_move):
        """Method to play the game in GUI using pygame

        Args:
            board (numpy.ndarray): Game board
            players (list): List of players
            is_valid_move (function): Move validator
            make_move (function): Makes move
            is_winning_move (function): Ckeck for winning move"""
        turn = random.randint(0, len(players) - 1)
        while True:
            col = (
                -1
            )  # Resets col, so game only move forward when a player(AI of human) makes a valid move
            self.draw(board)
            if players[turn].name == "AI":
                col = players[turn].get_move()
            else:
                for event in pygame.event.get():
                    if event.type == pygame.QUIT:
                        return None
                    if event.type == pygame.MOUSEMOTION:
                        self.draw_black_rec()
                        pygame.draw.circle(
                            self.screen,
                            C_LIST[players[turn].p_id],
                            (event.pos[0], int(SQUARESIZE / 2)),
                            RADIUS,
                        )
                    pygame.display.update()
                    if event.type == pygame.MOUSEBUTTONDOWN:
                        self.draw_black_rec()
                        col = int(math.floor(event.pos[0] / SQUARESIZE))
            if is_valid_move(col):
                row = make_move(col, players[turn].p_id)
                if is_winning_move(row, col):
                    self.blit(
                        " {} Wins!!".format(players[turn].name), players[turn].p_id
                    )
                    self.draw(board)
                    pygame.time.wait(3000)
                    return players[turn]
                turn = (turn + 1) % len(players)



"""Utility methods for project."""

import sys
from loguru import logger


def getVersion(file):
    """Method

    Args:
        file (str): File path(usually to the file version.txt)

    Returns:
        Version after reading from given file"""
    try:
        with open(file, "r") as f:
            return f.read()
    except FileNotFoundError:
        return ""


def recordGame(game):
    """Simple method to record games using loguru module by https://pypi.org/project/loguru/

    Args:
        game (ConnectNGame): Game object"""
    format_ = "<green>{time:DD-MM-YYYY at HH:mm:ss}</green>: <level>{message}</level>"
    record = """\nWinner: {winner}
        Players: {players}
        Game Sequence: {sequence}\n\n"""

    logger.add(
        "reports/GameRecords.log", colorize=False, rotation="50 MB", format=format_
    )
    logger.add(sys.stdout, colorize=True, format=format_)
    logger.info(
        record, winner=game.winner, players=game.players, sequence=game.sequence
    )


def dummy_method(*args, **kwargs):
    """Dummy method which does nothing
    Note: Accepts anything and everything"""
    pass



"""API for Connect-N game"""

# Flask module(s) and it's extentions
from flask import Flask
from flask_migrate import Migrate
from flask_sqlalchemy import SQLAlchemy

# Environment variable(s)
from .env import API_DB_NAME

db = SQLAlchemy()
migrate = Migrate()


def create_app(api_db_name=API_DB_NAME):
    """App factory function which runs setup for Flask app"""
    app = Flask(__name__)
    app.config["SQLALCHEMY_DATABASE_URI"] = "sqlite://" + api_db_name
    app.config["SQLALCHEMY_TRACK_MODIFICATIONS"] = False
    app.config["DEBUG"] = True
    db.init_app(app)
    migrate.init_app(app, db)
    return app


app = create_app()

# pylint: disable=unused-wildcard-import
from .auth import *
from .routes import *
from .tokens import *
from .error import *
from .db_model import *


# Helpful when using flask shell command
@app.shell_context_processor
def make_shell_context():
    return {"db": db, "User": User}



"""generate_data.py file, generates dataset for Deep Learning model, AI agent."""

# Python module(s)
import os
import random
import numpy as np

# Environment Variable(s)
from ..env import N, ROWS, COLUMNS

# Project Module(s)
from ..ai import AI
from ..player import Player
from ..connect_n import ConnectNGame


class GenerateData:
    """Generate Data class used to generate train data(in .npy format) for AI agent

    Args:
        out_dir (str): Output directory for training datast.
        file_name (str):
            File name for training dataset,
            (do not include extention in file name),
            extention .npy is automatically added
        game_runs (int): Number of game runs for training dataset.

    Raises:
        TypeError: When non-string type are passed for out_dir and file_name
        TypeError: When non-int type is passed for game_runs"""

    def __init__(
        self,
        out_dir="/dataset",
        file_name="dl_train_dataset",
        game_runs=10,
        n=N,
        num_rows=ROWS,
        num_cols=COLUMNS,
    ):
        if not isinstance(out_dir, str) or not isinstance(file_name, str):
            raise TypeError(
                "Type 'string' accepted for out_dir and file_name arguments, type {} and {} encountered".format(
                    type(out_dir), type(file_name)
                )
            )
        if not isinstance(game_runs, int):
            raise TypeError(
                "Type 'int' accepted for out_dir argument, type {} encountered".format(
                    type(game_runs)
                )
            )
        self.n = n
        self.num_rows = num_rows
        self.num_cols = num_cols
        self.game_runs = abs(game_runs)
        if not os.path.isdir(os.getcwd() + out_dir):
            os.mkdir(os.getcwd() + out_dir)
        self.out_file = os.getcwd() + "/" + out_dir + "/" + file_name + ".npy"

    def generate_save(self, want_return=False):
        """Generates dataset for Deep Learning Model trainning

        Args:
            want_return (bool): Returns generated dataset as numpy.ndarray if true

        Returns:
            numpy.ndarray : Generated dataset"""
        if self.game_runs == 0:
            raise MemoryError(
                "Dataset already generated, use another object for new dataset"
            )
        count = -1
        game_runs = self.game_runs
        dataset = list()
        game = ConnectNGame(ai=True, graphic=False)
        game.players.append(AI(game, 2))
        while game_runs:
            count += 1
            if count > (2 * self.game_runs):
                break
            winner = None
            board_states = [game.board]
            turn = random.getrandbits(1)
            game.reset()
            while not winner:
                if len(game.get_valid_moves()) == 0:
                    break
                turn = (turn + 1) % len(game.players)
                if random.getrandbits(1):
                    col = random.choice(game.get_valid_moves())
                elif random.getrandbits(1):
                    col = game.players[turn].get_move()
                else:
                    col = game.players[turn].greedy()
                row = game.make_move(col, turn + 1)
                board_states.append(game.board)
                if game.is_winning_move(row, col):
                    winner = "AI" if game.players[turn].p_id == 1 else "Not AI"
                    break
            if winner != "AI":
                continue
            game_runs -= 1
            dataset.append(np.array(board_states))
        np.save(self.out_file, np.array(dataset), allow_pickle=True)
        if want_return:
            return np.array(dataset)

    def load(self):
        """Loads generated dataset"""
        return np.load(self.out_file, allow_pickle=True)

    def __str__(self):
        """String representation of class object"""
        return "<class Generate Data>"



"""ai.py

File defining AI for the Connect-N Game"""

# Python module(s)
import math
import random

from .env import UNIT_SCORE, OFFSET, LARGE_VALUE, SMALL_VALUE

class AI:
    """class defining AI for the game
    Args:
        p_id (int): ID for AI(default: 1)
        game (ConnectNGame): Game object"""

    def __init__(self, game, p_id=1):
        """Instantiate Method"""
        self.p_id = p_id
        self.name = "AI"

        self.game = game
        self.n = game.n
        self.rows = game.rows
        self.cols = game.cols

    def greedy(self):
        """Simple method to fetch AI move, Uses scoring method to determine highest ranking move(greddy approach)

        Returns:
            int : Most optimal move (None if no legal moves)"""
        valid_loction = self.game.get_valid_moves()
        if len(valid_loction) == 0:  # No valid moves
            return None
        max_score = SMALL_VALUE
        best_move = random.choice(valid_loction)
        for col in valid_loction:
            temp_board = self.game.board.copy()
            self.game.make_move(col, self.p_id, temp_board)
            if max_score < self.score(temp_board, self.p_id):
                max_score = self.score(temp_board, self.p_id)
                best_move = col
        return best_move

    def get_move(self):
        """Fetches AI  move

        Returns:
            int : Most optimal move (None if no legal moves)"""
        if len(self.game.get_valid_moves()) == 0:
            return None
        # Check if AI winning, and win
        col = self.game.is_p_id_winning(self.p_id)
        if col:
            return col  # Retuen only if non None value returned

        try:  # Checks if next player is known, if not known finds out
            a = self.next_player.p_id
        except AttributeError:
            i = 0
            self.next_player = self.game.players[0]
            for p in self.game.players:
                if p.p_id == 1:
                    break
                i += 1
            if i < len(self.game.players):
                self.next_player = self.game.players[i + 1]
        else:
            del a

        # Check if next player is winning, and stop him/her
        col = self.game.is_p_id_winning(self.next_player.p_id)
        if col:
            return col

        return self.minmax(
            self.game.board.copy(), self.n + 1, -math.inf, math.inf, True
        )[0]

    def minmax(self, board, depth, alpha, beta, maximizingPlayer):
        """MinMax algortihm for AI implementation

        Args:
            board (numpu.ndarray): 2-D numpy array representing board in which game is being played
            depth (int): Depth upto which to look
            alpha (int): Parameter for alpha-beta purning
            beta (int): Parameter for alpha-beta purning
            maximizingPlayer (bool): Wheather maximizing player or not

        Returns:
            tuple : 2 length tuple, first element column, second score"""
        if self.game.is_p_id_winning(self.p_id, board):
            return (None, LARGE_VALUE)
        if self.game.is_p_id_winning(self.next_player.p_id, board):
            return (None, SMALL_VALUE)
        if len(self.game.get_valid_moves(board)) == 0:
            return (None, 0)
        if depth == 0:
            return (None, self.score(board, self.p_id))

        value = math.inf
        p_id = self.next_player.p_id
        column = random.choice(self.game.get_valid_moves(board))
        if maximizingPlayer:
            value = -math.inf
            p_id = self.p_id

        for col in self.game.get_valid_moves(board):
            b_copy = board.copy()
            row = self.game.make_move(col, p_id, b_copy)
            new_score = self.minmax(
                b_copy, depth - 1, alpha, beta, not maximizingPlayer
            )[1]
            if maximizingPlayer:
                if new_score > value:
                    value, column = new_score, col
                alpha = max(alpha, value)
            else:
                if new_score < value:
                    value, column = new_score, col
                beta = min(beta, value)
            if alpha >= beta:
                break
        del row
        return (column, value)

    def string_score(self, string, pid):
        """Calculates score for a player from given string

        Args:
            string (str): Input string
            pid (int): ID of player

        Returns:
            int : Score of the given string"""
        if str(pid) not in string:
            return 0
        else:
            pid = str(pid)
        score = OFFSET
        for begin in range(len(string) - self.n + 1):
            end = begin + self.n
            zeros = string[begin:end].count("0")
            me = string[begin:end].count(pid)
            other = self.n - zeros - me
            score += UNIT_SCORE * (zeros ** 4 + (me ** 4) * 3 + (other ** 4) * (-2.5))
        return score

    def score(self, board, pid):
        """Calculated Score of board for given player.
        Args:
            board (numpy.ndarray): A 2-D int numpy array representing current state of game
            pid (int): ID of player

        Returns:
            float : Score of the board for given pid"""
        return sum(
            self.string_score(string, pid) for string in self.game.get_strings(board)
        )

    def __str__(self):
        """String representation"""
        return "<class 'AI' {}>".format(self.p_id)



"""@author: Kartikei Mittal

@email: kartikeimittal@gmail.com

connect_n.py file

Contains main module of this repositry.

https://github.com/Kartikei-12/Connect-N"""


# Python module(s)
import os
import random
import numpy as np

# User module(s)
from .ai import AI
from .player import Player
from .utility import getVersion, recordGame, dummy_method

# Environment Variables
from .env import N, ROWS, COLUMNS


class ConnectNGame:
    """Main module class used Connect-N game.

    Args:
        ai (bool): Enable/Disable AI
        record (bool): Enable game recording.
        graphic (bool): GUI(true) or command line interface(false)
        n (int): Number of coins required in a line to win.
        num_rows (int): Number of rows
        num_col (int): Number of columns

    Raises:
        TypeError: Expected 'int' for num_rows, num_col, n
        TypeError: Expected 'bool' for ai, graphic, record
        ValueError: Argument [num_rows, num_col, n] needs to be positive
        ValueError: No winning combination possible in [num_rows, num_col, n]"""

    __version__ = "0.1d."

    def __init__(
        self, ai=True, record=False, graphic=False, n=N, num_rows=ROWS, num_col=COLUMNS
    ):
        """Instantiate function for class ConnectNGame"""

        # Checking 'bool' argument
        for var in (ai, graphic, record):
            if not isinstance(var, bool):
                raise TypeError("Expected 'bool' not {0}.".format(type(var)))
        # Checking 'int' argument
        for var in (n, num_rows, num_col):
            if not isinstance(var, int):
                raise TypeError("Expected 'int' not {0}".format(type(var)))
            elif var < 1:
                raise ValueError("All argument needs to be positive.")
            elif var < n:
                raise ValueError("No winning combination possible")

        PATH = os.path.dirname(os.path.realpath(__file__))
        self.__version__ += getVersion(PATH + "/version.txt")

        self.winner = None
        self.GUIUtil = None
        self.sequence = list()
        self.players = list()

        self.n = n
        self.rows = num_rows
        self.cols = num_col

        self.board = np.zeros((self.rows, self.cols), dtype=int)

        self.play = self.cmd_line
        self.record_game = dummy_method

        if ai:
            self.players.append(AI(self))
        if record:
            self.record_game = recordGame
        if graphic:
            from .pygame_utility import PygameUtility

            try:
                self.GUIUtil = PygameUtility(num_rows, num_col)
            except EnvironmentError as e:
                print(e)
            else:
                self.play = self.graphic

    def to_dict(self):
        """Provides a dictionary representaion of object"""
        object_dict = vars(self).copy()
        del object_dict["GUIUtil"]
        del object_dict["__version__"]
        del object_dict["play"]
        del object_dict["record_game"]
        del object_dict["players"]
        object_dict["board"] = list()
        for row in range(self.rows):
            for col in range(self.cols):
                object_dict["board"].append(int(self.board[row][col]))
        return object_dict

    def from_dict(self, game_dict):
        """Method to create game from dictioay

        Args:
            game_dict (dict): Dictionary containning state of the game"""
        self.winner = game_dict["winner"]
        self.sequence = game_dict["sequence"]

        self.n = game_dict["n"]
        self.rows = game_dict["rows"]
        self.cols = game_dict["cols"]

        self.board = np.reshape(game_dict["board"], (self.rows, self.cols))
        self.play = dummy_method
        self.record_game = dummy_method
        self.players.append(AI(self))

    def simulate(self, turn, sequence):
        """This method simulate a game from list of integers passed, each element representing column choosen by corresponding player to make move, Useful for trainning of AI.

        Note:
            * As soon as an invalid move is encountered game is aborted,
            * Index of cloumns belong to [0, self.cols), for example if self.cols is 6 column will to 0 to self.cols-1, limits included.

        Args:
            turn (int): Index of player from self.players list who gets the first turn
            sequence (list): List of moves by player in order of thir occurence in self.players list

        Returns:
            list : Refer self.sequence"""
        for col in sequence:
            if self.is_valid_move(col):
                row = self.make_move(col, self.players[turn].p_id)
            else:
                break
            if self.is_winning_move(row, col):
                self.winner = self.players[turn]
                break
            turn = (turn + 1) % len(self.players)
        return self.sequence

    def reset(self):
        """Resets th game for a new run.

        Note:
            Does NOT removes players from the game"""
        self.winner = None
        self.sequence = list()
        self.board = np.zeros((self.rows, self.cols), dtype=int)

    def add_player(self, p):
        """Method to add players to the game.

        Args:
            p (Player): player object

        Raises:
            TypeError: "Expected 'Player' not {0}".format(type(p))
            ValueError: '{} already in game.'.format(p)
        """
        if not isinstance(p, Player):
            raise TypeError("Expected 'Player' not {0}".format(type(p)))
        if p.p_id in [player.p_id for player in self.players]:
            raise ValueError("{} already in game.".format(p))
        self.players.append(p)

    def is_p_id_winning(self, p_id, board=None):
        """"""
        if board is None:
            board = self.board
        for col in self.get_valid_moves(board):
            board_copy = board.copy()
            row = self.make_move(col, p_id, board_copy)
            if self.is_winning_move(row, col, board_copy):
                return col
        return None

    def print_board(self):
        """Prints the board on console"""
        print(np.flip(self.board, 0))  # 0: Vertical flip

    def make_move(self, col, p_id, board=None):
        """Method to make move, returns row in which move was made

        Args:
            col (int): Column to insert coin in
            p_id (int): Id of player making the move
            board (numpy.ndarray): 2-D numpy array representing board in which game is being played

        Returns:
            int : Row in which move was made"""
        if board is None:
            board = self.board
            self.sequence.append(col)
        for row in range(self.rows):
            if board[row][col] == 0:
                board[row][col] = p_id
                return row
        return None

    def get_valid_moves(self, board=None):
        """Valid Moves

        Args:
            board (numpy.ndarray): 2-D numpy array representing board in which game is being played

        Returns:
            list : List of all valid moves"""
        if board is None:
            board = self.board
        return [i for i in range(self.cols) if self.is_valid_move(i, board)]

    def is_valid_move(self, col, board=None):
        """Check validity of move

        Args:
            col (int): Checks if move can be made in this column
            board (numpy.ndarray): 2-D numpy array representing board in which game is being played

        Returns:
            bool : True if valid move False otherwise"""
        if board is None:
            board = self.board
        if col < 0 or col >= self.cols or board[self.rows - 1][col] != 0:
            return False
        return True

    def is_winning_move(self, row, col, board=None):
        """Method to check for winning move,

        Note:
            Board is displayed in flipped position so,
            What appears to positive digonal is actually negative digonal,
            and vice versa.

        Args:
            row (int): Row in which last move was made
            col (int): Column in which last move was made
            board (numpy.ndarray): 2-D numpy array representing board in which game is being played

        Returns:
            bool : True if given position enabled a winning move"""
        if board is None:
            board = self.board
        if board[row][col] == 0:
            raise ValueError("Testing Empty slot in board.")
        desired_pat = "".join(str(board[row][col]) for i in range(self.n))

        # Horizontal Check
        if desired_pat in "".join(
            str(board[row][i])
            for i in range(max(0, col - self.n + 1), min(col + self.n, self.cols))
        ):
            return True

        # Vertical Check
        if desired_pat in "".join(
            str(board[i][col])
            for i in range(max(0, row - self.n + 1), min(row + self.n, self.rows))
        ):
            return True
        # Positive digonal Check
        if desired_pat in "".join(
            str(board[row + i][col + i])
            for i in range(-1 * (self.n - 1), self.n)
            if (row + i) >= 0
            and (col + i) >= 0
            and (row + i) < self.rows
            and (col + i) < self.cols
        ):
            return True
        # Negative digonal Check
        if desired_pat in "".join(
            str(board[row - i][col + i])
            for i in range(-1 * (self.n - 1), self.n)
            if (row - i) >= 0
            and (col + i) >= 0
            and (row - i) < self.rows
            and (col + i) < self.cols
        ):
            return True

    def get_strings(self, board=None):
        """Method generates all string of 'n' length from board

        Args:
            board (numpy.ndarray): A 2-D int numpy array representing current state of game

        Returns:
            list : A list of strings which may have any significance in the game"""
        strings = list()
        if board is None:
            board = self.board
        # Horizontal
        for i in range(self.rows):
            strings.append("".join(str(j) for j in board[i][...]))
        # Vertical
        for i in range(self.cols):
            strings.append("".join(str(j) for j in board[..., i]))
        # Positive Digonal Along Rows
        for i in range(self.n - 1, self.cols):
            strings.append(
                "".join(str(board[j][i - j]) for j in range(min(i + 1, self.rows)))
            )
        # Positive Digonal Along Columns
        for i in range(1, self.rows - self.n + 1):
            strings.append(
                "".join(
                    str(board[j + i][self.cols - j - 1]) for j in range(self.rows - i)
                )
            )
        # Negative Digonal Along Rows
        for i in range(self.cols - self.n + 1):
            strings.append(
                "".join(
                    str(board[j][i + j]) for j in range(min(self.rows, self.cols - i))
                )
            )
        # Negative Digonal Along Columns
        for i in range(1, self.rows - self.n + 1):
            strings.append("".join(str(board[i + j][j]) for j in range(self.rows - i)))
        return strings

    def cmd_line(self):
        """Method to play the game in command line"""
        turn = random.randint(0, len(self.players) - 1)
        while True:
            self.print_board()
            col = self.players[turn].get_move()
            if self.is_valid_move(col):
                row = self.make_move(col, self.players[turn].p_id)
                if self.is_winning_move(row, col):
                    self.winner = self.players[turn]
                    print("Winner: ", self.winner.name)
                    break
            else:
                print("Invalid move column already filled, aborting turn!")
            turn = (turn + 1) % len(self.players)
        self.record_game(self)

    def graphic(self):
        """Method to play the game in GUI using pygame"""
        if len(self.players) > 3:  # TODO: Remove magic number
            print("GUI currently only support 3 players.")
            self.cmd_line()
            return
        self.winner = self.GUIUtil.play(
            self.board,
            self.players,
            self.is_valid_move,
            self.make_move,
            self.is_winning_move,
        )
        self.record_game(self)

    def __str__(self):
        """Representation format:
            <class 'ConnectNGame', NUMBER_OF_ROWS NUMBER_OF_COLUMNS CONNECT_LENGTH>"""
        return "<class 'ConnectNGame', {0} {1} {2} >".format(
            self.rows, self.cols, self.n
        )



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="connect-n",
    version="0.0.5",
    author="Kartikei Mittal",
    author_email="kartikeimittal@gmail.com",
    description="A modified version of Connect-4 game implemented with AI using PyGame.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Kartikei-12/Connect-N",
    packages=setuptools.find_packages(exclude=("tests",)),
    install_requires=open("requirements.txt", "r").read().split("\n"),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
)



"""error.py Error managment for API"""

# Flask modules(s)
from flask import jsonify
from werkzeug.http import HTTP_STATUS_CODES

from .app import app, db
from .utility import compile_response


def bad_request(text, status_code):
    """Response generator for bad request to API"""
    payload = {
        "error": HTTP_STATUS_CODES.get(status_code, "Unknown error"),
        "message": text,
    }
    response = jsonify(compile_response(description=payload))
    response.status_code = status_code
    return response


@app.errorhandler(404)
def not_found_error(error):
    """"""
    return bad_request("Not found", 404)


@app.errorhandler(500)
def internal_error(error):
    """"""
    db.session.rollback()
    return bad_request("Internal Server Error", 500)
